/*
 Copyright (C) 2015 Andrey Tsyganov, UlSPU. All rights reserved.

 Contact information
 -------------------
 Andrey Tsyganov, Ulyanovsk State Pedagogical University, Laboratory of Mathematical Modeling
 Web    :  http://lmm.ulspu.ru
 e-mail :  andrew.tsyganov@gmail.com 
*/


#ifndef GA_H
#define GA_H


// Remember about memory limitations!
// http://www.arduino.cc/en/Tutorial/Memory
// http://playground.arduino.cc/Main/CorruptArrayVariablesAndMemory


//#define GA_DEBUG // Uncomment for debugging, comment to save memory and avoid memory corruption


#if defined(ARDUINO) && ARDUINO >= 100
#include "Arduino.h" // Arduino IDE >= v1.0
#else
#include "WProgram.h"
#endif


#ifdef GA_DEBUG

int freeRam() 
{
    extern int __heap_start, *__brkval; 
    int v; 
    return (int) &v - (__brkval == 0 ? (int) &__heap_start : (int) __brkval); 
}

#endif


enum stop_t {GEN_LIM, FITNESS_LIM, TIME_LIM};


typedef byte          pop_t;     // Type to store population size, elite count, etc.
typedef unsigned int  gen_t;     // Type to store number of generations. You may change it to any integer type 
typedef unsigned long time_t;    // Type to store time


const pop_t MAX_POP_SIZE = 25;


// GA is an abstract class from which you have to derive your own genetic algorithm.
// Template parameter TFitness is a float or integer type returned by fitness function.
// Template parameter TIndiv is a class provided by user. This class describes individuals of the population (see example sketches).
template <class TFitness, class TIndiv> 
class GA
{
public:
    // Parameters initialization
    GA()
    {
        pop_size_       = 20;
        stop_cond_      = GEN_LIM;    
        gen_lim_        = 100;
        fitness_lim_    = 0;
        time_lim_       = 0;
        elite_count_    = 2;
        dead_count_     = 2;
        crossover_frac_ = 0.8;
    };

    // Parameters setters
    void setPopSize(pop_t pop_size);
    void setEliteCount(pop_t elite_count);
    void setDeadCount(pop_t dead_count);
    void setCrossoverFrac(float crossover_frac);
    void setStopCond(stop_t stop_cond)       { stop_cond_ = stop_cond; };
    void setGenLim(gen_t gen_lim)            { gen_lim_ = gen_lim; };
    void setFitnessLim(TFitness fitness_lim) { fitness_lim_ = fitness_lim; };
    void setTimeLim(time_t time_lim)         { time_lim_ = time_lim; };

    // Parameters getters
    pop_t     getPopSize()       const { return pop_size_; };
    pop_t     getEliteCount()    const { return elite_count_; };
    pop_t     getDeadCount()     const { return dead_count_; };
    float     getCrossoverFrac() const { return crossover_frac_; };
    stop_t    getStopCond()      const { return stop_cond_; };
    gen_t     getGenLim()        const { return gen_lim_; };
    TFitness  getFitnessLim()    const { return fitness_lim_; };
    time_t    getTimeLim()       const { return time_lim_; };
    
    // Other getters
    gen_t     getGen()           const { return gen_; };
    TIndiv    getIndiv(pop_t n)  const;

    // Main method
    void run();

#ifdef GA_DEBUG
	// Debugging methods
	void printParam() const;
	void printState() const;
	void printProgress() const;
#endif

protected:
	// Implement these pure virtual problem-dependent methods in derived class
	virtual TIndiv   generate() const = 0;                                              // Generate (random) individual
	virtual TFitness fitness(TIndiv const& indiv) const = 0;                            // Evaluate fitness
	virtual TIndiv   crossover(TIndiv const& parent1, TIndiv const& parent2) const = 0; // Perform crossover
	virtual TIndiv   mutation(TIndiv const& indiv) const = 0;                           // Perform mutation
	virtual void     print(TIndiv const& indiv) const = 0;                              // Print individual

	// GA parameters
    pop_t     pop_size_;
    pop_t     elite_count_;
    pop_t     dead_count_;
    float     crossover_frac_;
    stop_t    stop_cond_;
    gen_t     gen_lim_;
    TFitness  fitness_lim_;
    time_t    time_lim_;

    // Population
    TIndiv population_[MAX_POP_SIZE];
    TIndiv new_population_[MAX_POP_SIZE];
    
    // Individuals fitnesses
    TFitness fitnesses_[MAX_POP_SIZE];

    // GA variables
    time_t start_time_;
    gen_t  gen_;
    pop_t  crossover_count_, mutation_count_, alive_count_;

    // Stop condition checker
    boolean stopCondition();
    void sortPopulation();
};


template <class TFitness, class TIndiv> 
void GA<TFitness, TIndiv>::setPopSize(pop_t pop_size)
{
    if ((pop_size <= MAX_POP_SIZE) && (pop_size >= elite_count_ + dead_count_))
    {
        pop_size_ = pop_size;
        return;
    }

#ifdef GA_DEBUG
    Serial.println(F("Incorrect population size!"));
#endif
};


template <class TFitness, class TIndiv> 
void GA<TFitness, TIndiv>::setEliteCount(pop_t elite_count)
{
    if (elite_count + dead_count_ <= pop_size_)
    {
        elite_count_ = elite_count;
        return;
    }

#ifdef GA_DEBUG
    Serial.println(F("Incorrect elite counter!"));
#endif
};


template <class TFitness, class TIndiv> 
void GA<TFitness, TIndiv>::setDeadCount(pop_t dead_count)
{
    if (dead_count + elite_count_ <= pop_size_)
    {
        dead_count_ = dead_count;
        return;
    }

#ifdef GA_DEBUG
    Serial.println(F("Incorrect dead counter!"));
#endif
};


template <class TFitness, class TIndiv> 
void GA<TFitness, TIndiv>::setCrossoverFrac(float crossover_frac)
{
    if ((crossover_frac >= 0) && (crossover_frac <= 1))
    {
        crossover_frac_ = crossover_frac;
        return;
    };
    
#ifdef GA_DEBUG
    Serial.println(F("Incorrect crossover fraction!"));
#endif
};


template <class TFitness, class TIndiv> 
TIndiv GA<TFitness, TIndiv>::getIndiv(pop_t n) const
{ 
    if (n < pop_size_)
    {
        return population_[n];
    }
	else
	{
#ifdef GA_DEBUG
		Serial.println(F("Incorrect index!"));
#endif
		return population_[pop_size_ - 1];
	}
}


template <class TFitness, class TIndiv> 
void GA<TFitness, TIndiv>::run()
{
    start_time_      = micros();
    crossover_count_ = (pop_t) (((float) (pop_size_ - elite_count_)) * crossover_frac_);
    mutation_count_  = pop_size_ - elite_count_ - crossover_count_;
    alive_count_     = pop_size_ - dead_count_;

    gen_ = 0;

    // Randomly generate initial population and evaluate individuals fitnesses
    for (pop_t i = 0; i < pop_size_; i++)
    {
        population_[i] = generate();
        fitnesses_[i]  = fitness(population_[i]);
    }

    // Dead individuals are now at the end
    sortPopulation();

#ifdef GA_DEBUG
    printProgress();
#endif

    while (!stopCondition())
    {
        gen_++;

        // Apply crossover on alive individuals to produce crossover children
        for (pop_t i = 0; i < crossover_count_; i++)
        {
            pop_t ind1 = random(0, alive_count_);
            pop_t ind2 = random(0, alive_count_);

            if (ind1 == ind2)
                (ind1 == 0 ? ind2++ : ind2--); // alive_count ???

            if (fitnesses_[ind1] < fitnesses_[ind2])
                new_population_[elite_count_ + i] = crossover(population_[ind1], population_[ind2]);
            else
                new_population_[elite_count_ + i] = crossover(population_[ind2], population_[ind1]);
        }

        // Apply mutation on alive individuals to produce mutation children
        for (pop_t i = 0; i < mutation_count_; i++)
            new_population_[elite_count_ + crossover_count_ + i] = mutation(population_[random(0, alive_count_)]);

        // Update population (elite individuals are already at the beginning)
        for (pop_t i = elite_count_; i < pop_size_; i++)
        {
            population_[i] = new_population_[i];
            fitnesses_[i]  = fitness(population_[i]);
        }

        sortPopulation();

#ifdef GA_DEBUG
        printProgress();
#endif
    }
};


#ifdef GA_DEBUG
template <class TFitness, class TIndiv> 
void GA<TFitness, TIndiv>::printParam() const
{
    Serial.print(F("pop_size_ = "));
    Serial.println(pop_size_);
    Serial.print(F("stop_cond_ = "));
    switch (stop_cond_)
    {
        case GEN_LIM:     Serial.println(F("GEN_LIM")); break;
        case FITNESS_LIM: Serial.println(F("FITNESS_LIM")); break;
        case TIME_LIM:    Serial.println(F("TIME_LIM")); break;
    }
    Serial.print(F("gen_lim_ = "));
    Serial.println(gen_lim_);
    Serial.print(F("fitness_lim_ = "));
    Serial.println(fitness_lim_);
    Serial.print(F("time_lim_ = "));
    Serial.println(time_lim_);
    Serial.print(F("elite_count_ = "));
    Serial.println(elite_count_);
    Serial.print(F("dead_count_ = "));
    Serial.println(dead_count_);
    Serial.print(F("crossover_frac_ = "));
    Serial.println(crossover_frac_);
}


template <class TFitness, class TIndiv> 
void GA<TFitness, TIndiv>::printState() const
{
    Serial.println();
    Serial.print(F("Free RAM (bytes): "));
    Serial.println(freeRam());
    Serial.print(F("gen_ = "));
    Serial.println(gen_);
    Serial.print(F("elite_count_ = "));
    Serial.println(elite_count_);
    Serial.print(F("crossover_count_ = "));
    Serial.println(crossover_count_);
    Serial.print(F("mutation_count_ = "));
    Serial.println(mutation_count_);
    Serial.print(F("alive_count_ = "));
    Serial.println(alive_count_);

    Serial.println(F("population_:"));
    
    for (pop_t i = 0; i < pop_size_; i++)
    {
        print(population_[i]);
        Serial.print(F("    "));
    }
    Serial.println();
    
    Serial.println(F("fitnesses_:"));
    for (pop_t i = 0; i < pop_size_; i++)
    {
        Serial.print(fitnesses_[i]);
        Serial.print(F("    "));
    }
    Serial.println();

    Serial.println(F("new_population_:"));
    
    for (pop_t i = 0; i < pop_size_; i++)
    {
        print(new_population_[i]);
        Serial.print(F("    "));
    }
    Serial.println();
}


template <class TFitness, class TIndiv> 
void GA<TFitness, TIndiv>::printProgress() const
{
    Serial.print(F("gen_ = "));
    Serial.print(gen_);
    Serial.print(F("    best = "));
    print(population_[0]);
    Serial.print(F("    fitness = "));
    Serial.println(fitnesses_[0]);
}
#endif


template <class TFitness, class TIndiv> 
boolean GA<TFitness, TIndiv>::stopCondition()
{
    switch (stop_cond_)
    {
        case GEN_LIM:     return (gen_ >= gen_lim_ ? true : false); 
        case FITNESS_LIM: return (fitnesses_[0] <= fitness_lim_ ? true : false);
        case TIME_LIM:    return ((micros() - start_time_) >= time_lim_ ? true : false);
    }
};


template <class TFitness, class TIndiv> 
void GA<TFitness, TIndiv>::sortPopulation()
{
    for(pop_t i = 0; i < pop_size_ - 1; i++)
        for(pop_t j = 0; j < pop_size_ - (i + 1); j++)
            if (fitnesses_[j] > fitnesses_[j + 1])
            {
                TFitness temp_fit = fitnesses_[j];
                TIndiv temp_indiv = population_[j];

                fitnesses_[j]  = fitnesses_[j + 1];
                population_[j] = population_[j + 1];

                fitnesses_[j + 1]  = temp_fit;
                population_[j + 1] = temp_indiv;
            }
}
 
#endif
