/*
 Copyright (C) 2015 Andrey Tsyganov, UlSPU. All rights reserved.

 Contact information
 -------------------
 Andrey Tsyganov, Ulyanovsk State Pedagogical University, Laboratory of Mathematical Modeling
 Web    :  http://lmm.ulspu.ru
 e-mail :  andrew.tsyganov@gmail.com 
*/

#include "GA.h"
#include "math.h"

const int   dim = 4;         // Number of dimensions
const float lbound  = -5.12; // Lower bound for each gene
const float ubound  =  5.12; // Upper bound for each gene
const float max_bias = 0.5;  // Maximum bias of the mutated gene

// some constants to speedup calculations
const float k  = 2 * M_PI;                          // constant 2 * Pi coefficient in fitness function
const float s0 = 10 * dim;                          // constant 10 * dim term in fitness fuction
const float gfactor = (ubound - lbound) / RAND_MAX; // constant mapping coefficient for random generator in generate() 
const float mfactor = 2.0 * max_bias / RAND_MAX;    // constant mapping coefficient for random generator in mutation()

class GAIndiv
{
    friend class MyGA;
private:
    float x_[dim]; // genes 
};

class MyGA: public GA<float, GAIndiv>
{
public:
    virtual GAIndiv generate() const;
    virtual float   fitness(GAIndiv const& indiv) const;
    virtual GAIndiv crossover(GAIndiv const& parent1, GAIndiv const& parent2) const;
    virtual GAIndiv mutation(GAIndiv const& indiv) const;
    virtual void    print(GAIndiv const& indiv) const;
};

GAIndiv MyGA::generate() const
{
    GAIndiv indiv;
    for (int i = 0; i < dim; i++) 
        indiv.x_[i] = lbound + gfactor * ((float) random(RAND_MAX));
    
    return indiv;    
};

float MyGA::fitness(GAIndiv const& indiv) const
{
    float sum = s0;
    for (int i = 0; i < dim; i++)
        sum += indiv.x_[i] * indiv.x_[i] - 10 * cos(k * indiv.x_[i]);

    return sum;
}

GAIndiv MyGA::crossover(GAIndiv const& parent1, GAIndiv const& parent2) const
{
    // Scattered crossover
    GAIndiv child;
    for (int i = 0; i < dim; i++)
        // Better (first) parent contributes two times more genes
        child.x_[i] = !random(3) ? parent2.x_[i] : parent1.x_[i];

    return child;
};

GAIndiv MyGA::mutation(GAIndiv const& indiv) const
{
    GAIndiv child;
    for (int i = 0; i < dim; i++)
    {
        child.x_[i] +=  - max_bias + mfactor * ((float) random(RAND_MAX));
            
        if (child.x_[i] < lbound) 
            child.x_[i] = lbound;

        if (child.x_[i] > ubound) 
            child.x_[i] = ubound;
    }

    return child;
};

void MyGA::print(GAIndiv const& indiv) const
{
    Serial.print("(");
    for (int i = 0; i < dim; i++)
    {
        Serial.print(indiv.x_[i], 8);
        if (i < dim - 1) 
            Serial.print("; ");
    }
    Serial.print(")");
};



MyGA ga;
GAIndiv indiv;

time_t start_time, end_time;
int seed;

void setup() 
{
  Serial.begin(115200);
  
  ga.setPopSize(15);
  ga.setEliteCount(1);
  ga.setCrossoverFrac(0.7);
  
  //ga.setStopCond(GEN_LIM);
  //ga.setGenLim(200);
  
  ga.setStopCond(FITNESS_LIM);
  ga.setFitnessLim(0.001);
  
  //ga.setStopCond(TIME_LIM);
  //ga.setTimeLim(4000000); // 4 sec
  
  //Serial.println("GA Parameters");
  //ga.printParam();
  
  //seed = analogRead(0);
  seed = 0;
}


void loop() 
{
  randomSeed(seed);
  Serial.println();
  Serial.print("Seed: ");
  Serial.println(seed);

  Serial.println("Start GA");
  start_time = micros();
  ga.run();
  end_time   = micros();
  Serial.println("End GA");

  Serial.print("Elapsed time: ");
  Serial.print((float) (end_time - start_time) / 1000000, 6);
  Serial.println(" second(s)");
  Serial.print("Generations: ");
  Serial.println(ga.getGen());
  indiv = ga.getIndiv(0);
  Serial.print("Best solution: ");
  ga.print(indiv);
  Serial.println();
  Serial.print("Fitness: ");
  Serial.println(ga.fitness(indiv), 8);
  
  seed++;
}
