/*
 Copyright (C) 2015 Andrey Tsyganov, UlSPU. All rights reserved.

 Contact information
 -------------------
 Andrey Tsyganov, Ulyanovsk State Pedagogical University, Laboratory of Mathematical Modeling
 Web    :  http://lmm.ulspu.ru
 e-mail :  andrew.tsyganov@gmail.com 
*/


#ifndef SA_H
#define SA_H


// Remember about memory limitations!
// http://www.arduino.cc/en/Tutorial/Memory
// http://playground.arduino.cc/Main/CorruptArrayVariablesAndMemory


//#define SA_DEBUG // Uncomment for debugging, comment to save memory and avoid memory corruption


#if defined(ARDUINO) && ARDUINO >= 100
#include "Arduino.h" // Arduino IDE >= v1.0
#else
#include "WProgram.h"
#endif


#include "math.h"


#ifdef SA_DEBUG
int freeRam() 
{
    extern int __heap_start, *__brkval; 
    int v; 
    return (int) &v - (__brkval == 0 ? (int) &__heap_start : (int) __brkval); 
}
#endif


enum stop_t {STEPS_LIM, COST_LIM, TEMP_LIM, TIME_LIM, NO_ACCEPTANCE};
enum cool_t {EXP, LINEAR, FAST, BOLTZ}; // http://www.mathworks.com/help/gads/how-simulated-annealing-works.html


typedef unsigned int  count_t; // Type to store counters. You may change it to any integer type. 
typedef unsigned long time_t;  // Type to store time.


// SA is an abstract class from which you have to derive your own simulated annealing algorithm.
// Template parameter TCost is a float or integer type returned by cost function.
// Template parameter TSolution is a class provided by user. This class describes feasible solution of the problem (see example sketches).
template <class TCost, class TSolution> 
class SA
{
public:
    // Constructor
    SA()
    {
        stop_cond_    = STEPS_LIM;
        steps_lim_    = 10;
        cost_lim_     = 0;
        temp_lim_     = 0;
        time_lim_     = 0;
        step_len_     = 50;
        initial_temp_ = 1;
        initial_prob_ = 0.9;
        cooling_      = EXP;
        cooling_rate_ = 0.95; 
    };

    // Parameters setters
    void setStopCond(stop_t stop_cond)      { stop_cond_ = stop_cond; };
    void setStepsLim(count_t steps_lim)     { steps_lim_ = steps_lim; };
    void setCostLim(TCost cost_lim)         { cost_lim_ = cost_lim; };
    void setTempLim(float temp_lim)         { temp_lim_ = temp_lim; };
    void setTimeLim(time_t time_lim);
    void setStepLen(count_t step_len)       { step_len_ = step_len; };
    void setInitialTemp(float initial_temp);
    void setInitialProb(float initial_prob);
    void setCooling(cool_t cooling)         { cooling_  = cooling; };
    void setCoolingRate(float cooling_rate) { cooling_rate_ = cooling_rate; };

    // Parameters getters
    stop_t  getStopCond()    const { return stop_cond_; };
    count_t getStepsLim()    const { return steps_lim_; };
    TCost   getCostLim()     const { return cost_lim_; };
    float   getTempLim()     const { return temp_lim_; };
    time_t  getTimeLim()     const { return time_lim_; };
    count_t getStepLen()     const { return step_len_; };
    float   getInitialTemp() const { return initial_temp_; };
    cool_t  getCooling()     const { return cooling_; };
    float   getCoolingRate() const { return cooling_rate_; };

    // Other getters
    TSolution getBest()  const { return best_; }; // Get best solution

    // Main method
    void run();

#ifdef SA_DEBUG
    // Print methods (for debugging purposes)
    void printParam() const;
    void printState() const;
    void printProgress() const;
#endif

protected:
    // Implement these pure virtual problem-dependent methods in derived class
    virtual TSolution generate() const = 0;                           // Generate (random) solution
    virtual TCost     cost(TSolution const& solution) const = 0;      // Evaluate cost
    virtual TSolution neighbour(TSolution const& solution) const = 0; // Generate (random) neighbour
    virtual void      print(TSolution const& solution) const = 0;     // Print solution

    // SA parameters
    stop_t  stop_cond_;
    count_t steps_lim_;
    TCost   cost_lim_;
    float   temp_lim_;
    time_t  time_lim_;
    count_t step_len_;
    float   initial_temp_;
    float   initial_prob_;
    cool_t  cooling_;
    float   cooling_rate_;
 
    // SA variables
    time_t    start_time_;
    float     temp_;
    count_t   step_, step_iter_;
    TSolution current_, new_, best_;
    TCost     current_cost_, new_cost_, best_cost_;
    boolean   no_acceptance_;

    boolean stopCondition(); // Stop condition checker
    boolean acceptNew();     // Acceptance function
    void updateTemp();       // Temperature updater
};


template <class TCost, class TSolution>
void SA<TCost, TSolution>::setInitialTemp(float initial_temp)
{
  if (initial_temp > 0)
  {
    initial_temp_ = initial_temp;
    return;
  };

#ifdef SA_DEBUG
    Serial.println(F("Incorrect initial temperature!"));
#endif
};


template <class TCost, class TSolution> 
void SA<TCost, TSolution>::setInitialProb(float initial_prob) 
{ 
    if ((initial_prob > 0) && (initial_prob < 0.99))
    {
        initial_prob_ = initial_prob;
        return;
    };
    
#ifdef SA_DEBUG
    Serial.println(F("Incorrect initial probability!"));
#endif
};


template <class TCost, class TSolution>
void SA<TCost, TSolution>::setTimeLim(time_t time_lim)
{
  if (time_lim >= 0)
  {
    time_lim_ = time_lim;
    return;
  };

#ifdef SA_DEBUG
    Serial.println(F("Incorrect time limit!"));
#endif
};

template <class TCost, class TSolution> 
void SA<TCost, TSolution>::run()
{
    start_time_ = micros();

    // Calculate initial temperature
    float max_delta = 0;
    for (int i = 0; i < 20; i++)
    {
      float delta = abs(cost(generate()) - cost(generate()));
      if (delta > max_delta)
          max_delta = delta;
    }
  
    float est_t0 = - max_delta / log(initial_prob_); 
    initial_temp_ = max(initial_temp_, est_t0);
    temp_  = initial_temp_;
    
    // Calculate initial solution
    step_ = 0;
    no_acceptance_ = false;

    current_      = generate();
    current_cost_ = cost(current_);

    best_         = current_;
    best_cost_    = current_cost_;

#ifdef SA_DEBUG
        printProgress();
        //printState();
#endif

    // Main loop
    while (!stopCondition())
    {
        step_++;
        no_acceptance_ = true;
        for (step_iter_ = 1; step_iter_ <= step_len_; step_iter_++)
        {
            new_ = neighbour(current_);
            new_cost_ = cost(new_);
           
            if (acceptNew())
            {
                no_acceptance_ = false;
                current_ = new_;
                current_cost_ = new_cost_;

                if (current_cost_ < best_cost_)
                {
                    best_ = current_;
                    best_cost_ = current_cost_;
                }
            }
        }

#ifdef SA_DEBUG
        printProgress();
        //printState();
#endif

        updateTemp();
    }
};


template <class TCost, class TSolution> 
boolean SA<TCost, TSolution>::stopCondition()
{
    switch (stop_cond_) 
    {
        case STEPS_LIM:      return (step_ > steps_lim_ ? true : false); 
        case COST_LIM:       return (best_cost_ <= cost_lim_ ? true : false);
        case TEMP_LIM:       return (temp_ <= temp_lim_ ? true : false);
        case TIME_LIM:       return ((micros() - start_time_) >= time_lim_ ? true : false);
        case NO_ACCEPTANCE:  return (no_acceptance_ ? true : false);
    }
};


template <class TCost, class TSolution> 
boolean SA<TCost, TSolution>::acceptNew()
{
    return (new_cost_ < current_cost_ ? true : exp(-(new_cost_ - current_cost_) / temp_) > ((float) random(RAND_MAX) / (float) RAND_MAX));
};


template <class TCost, class TSolution> 
void SA<TCost, TSolution>::updateTemp()
{
    switch (cooling_)
    {
        case EXP:     temp_ *= cooling_rate_; break;
        case LINEAR:  temp_ -= cooling_rate_; break;
        case FAST:    temp_  = initial_temp_ / (1 + step_); break;
        case BOLTZ:   temp_  = initial_temp_ / log(1 + step_); break;
    }
};


#ifdef SA_DEBUG
template <class TCost, class TSolution> 
void SA<TCost, TSolution>::printParam() const
{
    switch (stop_cond_)
    {
        case STEPS_LIM:     Serial.println(F("STEPS_LIM")); break;
        case COST_LIM:      Serial.println(F("COST_LIM")); break;
        case TEMP_LIM:      Serial.println(F("TEMP_LIM")); break;
        case TIME_LIM:      Serial.println(F("TIME_LIM")); break;
        case NO_ACCEPTANCE: Serial.println(F("NO_ACCEPTANCE")); break;
    }

    Serial.print(F("steps_lim_ = "));
    Serial.println(steps_lim_);
    Serial.print(F("cost_lim_ = "));
    Serial.println(cost_lim_);
    Serial.print(F("temp_lim_ = "));
    Serial.println(temp_lim_, 6);
    Serial.print(F("time_lim_ = "));
    Serial.println(time_lim_);
    Serial.print(F("step_len_ = "));
    Serial.println(step_len_);
    Serial.print(F("initial_temp_ = "));
    Serial.println(initial_temp_, 6);
    Serial.print(F("initial_prob_ = "));
    Serial.println(initial_prob_, 6);

    switch (cooling_)
    {
        case EXP:    Serial.println(F("EXP")); break;
        case LINEAR: Serial.println(F("LINEAR")); break;
        case FAST:   Serial.println(F("FAST")); break;
        case BOLTZ:  Serial.println(F("BOLTZ")); break;
    }

    Serial.print(F("cooling_rate_ = "));
    Serial.println(cooling_rate_, 6);
}


template <class TCost, class TSolution> 
void SA<TCost, TSolution>::printState() const
{
    Serial.println();
    Serial.print(F("Free RAM (bytes): "));
    Serial.println(freeRam());
    Serial.print(F("temp_ = "));
    Serial.println(temp_, 6);
    Serial.print(F("step_ = "));
    Serial.println(step_);
    //Serial.print(F("step_iter_ = "));
    //Serial.println(step_iter_);
    Serial.println(F("current_:"));
    print(current_);
    Serial.println();
    Serial.print(F("current_cost_ = "));
    Serial.println(current_cost_);
    Serial.println(F("new_:"));
    print(new_);
    Serial.println();
    Serial.print(F("new_cost_ = "));
    Serial.println(new_cost_);
    Serial.println(F("best_:"));
    print(best_);
    Serial.println();
    Serial.print(F("best_cost_ = "));
    Serial.println(best_cost_);
}


template <class TCost, class TSolution> 
void SA<TCost, TSolution>::printProgress() const
{
    Serial.print(F("temp_ = "));
    Serial.print(temp_, 8);
    Serial.print(F("    step_ = "));
    Serial.print(step_);
    Serial.print(F("    step_iter_ = "));
    Serial.print(step_iter_);
    Serial.print(F("    best_ = "));
    print(best_);
    Serial.print(F("    cost_ = "));
    Serial.println(best_cost_);
}
#endif

#endif
