/*
 Copyright (C) 2015 Andrey Tsyganov, UlSPU. All rights reserved.

 Contact information
 -------------------
 Andrey Tsyganov, Ulyanovsk State Pedagogical University, Laboratory of Mathematical Modeling
 Web    :  http://lmm.ulspu.ru
 e-mail :  andrew.tsyganov@gmail.com 
*/

#include "SA.h"
#include "math.h"

const int   dim = 3;                // Number of dimensions
const int   change_prob_rate = dim; // One out of change_prob_rate coordinates is changed randomly and the others toward the best_ in neighbour()
const float lbound  = -5.12;        // Lower bound for each coordinate
const float ubound  =  5.12;        // Upper bound for each coordinate
const float max_bias = 0.5;         // Maximum bias of the coordinate

// some constants to speedup calculations
const float k  = 2 * M_PI;                          // constant 2 * Pi coefficient in cost function
const float s0 = 10 * dim;                          // constant 10 * dim term in cost fuction
const float gfactor = (ubound - lbound) / RAND_MAX; // constant mapping coefficient for random generator in generate() 
const float mfactor = 2.0 * max_bias / RAND_MAX;    // constant mapping coefficient for random generator in neighbour()

class SASolution
{
    friend class MySA;
private:
    float x_[dim]; // genes 
};

class MySA: public SA<float, SASolution>
{
public:
    virtual SASolution generate() const;
    virtual float   cost(SASolution const& solution) const;
    virtual SASolution neighbour(SASolution const& solution) const;
    virtual void    print(SASolution const& solution) const;
};

SASolution MySA::generate() const
{
    SASolution solution;
    for (int i = 0; i < dim; i++) 
        solution.x_[i] = lbound + gfactor * ((float) random(RAND_MAX));
    
    return solution;    
};

float MySA::cost(SASolution const& solution) const
{
    float sum = s0;
    for (int i = 0; i < dim; i++)
        sum += solution.x_[i] * solution.x_[i] - 10 * cos(k * solution.x_[i]);

    return sum;
}

SASolution MySA::neighbour(SASolution const& solution) const
{
    SASolution neighb;
    for (int i = 0; i < dim; i++)
    {
          if (!random(change_prob_rate))
          {
              neighb.x_[i] +=  - max_bias + mfactor * ((float) random(RAND_MAX));
          }
          else
          {
              float alpha = (float) random(RAND_MAX) / (float) RAND_MAX;
              neighb.x_[i] =  alpha * neighb.x_[i] + (1 - alpha) * best_.x_[i];
          }
            
          if (neighb.x_[i] < lbound) 
              neighb.x_[i] = lbound;

          if (neighb.x_[i] > ubound) 
              neighb.x_[i] = ubound;
    }

    return neighb;
};

void MySA::print(SASolution const& solution) const
{
    Serial.print("(");
    for (int i = 0; i < dim; i++)
    {
        Serial.print(solution.x_[i], 8);
        if (i < dim - 1) 
            Serial.print("; ");
    }
    Serial.print(")");
};



MySA sa;
SASolution solution;

time_t start_time, end_time;
int seed;

void setup() 
{
  Serial.begin(115200);
  
  sa.setStopCond(COST_LIM);
  sa.setCostLim(0.01);
  //sa.setStopCond(STEPS_LIM);
  //sa.setStepsLim(1000);
  //sa.setStopCond(NO_ACCEPTANCE);
  //sa.setCooling(BOLTZ);
  sa.setCooling(FAST);
  
  //Serial.println("SA Parameters");
  //sa.printParam();
  
  //seed = analogRead(0);
  seed = 0;
}


void loop() 
{
  randomSeed(seed);
  Serial.println();
  Serial.print("Seed: ");
  Serial.println(seed);

  Serial.println("Start SA");
  start_time = micros();
  sa.run();
  end_time   = micros();
  Serial.println("End SA");

  Serial.print("Elapsed time: ");
  Serial.print((float) (end_time - start_time) / 1000000, 6);
  Serial.println(" second(s)");
  solution = sa.getBest();
  Serial.print("Best solution: ");
  sa.print(solution);
  Serial.println();
  Serial.print("Cost: ");
  Serial.println(sa.cost(solution));
  
  seed++;
}
