/*
 Copyright (C) 2015 Andrey Tsyganov, UlSPU. All rights reserved.

 Contact information
 -------------------
 Andrey Tsyganov, Ulyanovsk State Pedagogical University, Laboratory of Mathematical Modeling
 Web    :  http://lmm.ulspu.ru
 e-mail :  andrew.tsyganov@gmail.com 
*/


#ifndef SHC_H
#define SHC_H


// Remember about memory limitations!
// http://www.arduino.cc/en/Tutorial/Memory
// http://playground.arduino.cc/Main/CorruptArrayVariablesAndMemory


//#define SHC_DEBUG // Uncomment for debugging, comment to save memory and avoid memory corruption


#if defined(ARDUINO) && ARDUINO >= 100
#include "Arduino.h" // Arduino IDE >= v1.0
#else
#include "WProgram.h"
#endif


#include "math.h"


#ifdef SHC_DEBUG

int freeRam() 
{
    extern int __heap_start, *__brkval; 
    int v; 
    return (int) &v - (__brkval == 0 ? (int) &__heap_start : (int) __brkval); 
}

#endif


enum stop_t {STEPS_LIM, COST_LIM, TEMP_LIM, TIME_LIM, NO_ACCEPTANCE};


typedef unsigned int  count_t; // Type to store counters. You may change it to any integer type. 
typedef unsigned long time_t;  // Type to store time.


// SHC is an abstract class from which you have to derive your own stochastic hill climbing algorithm.
// Template parameter TCost is a float or integer type returned by cost function.
// Template parameter TSolution is a class provided by user. This class describes feasible solution of the problem (see example sketches).
template <class TCost, class TSolution> 
class SHC
{
public:
    // Constructor
    SHC()
    {
        stop_cond_    = STEPS_LIM;
        steps_lim_    = 10;
        cost_lim_     = 0;
        time_lim_     = 0;
    };

    // Parameters setters
    void setStopCond(stop_t stop_cond)      { stop_cond_ = stop_cond; };
    void setStepsLim(count_t steps_lim)     { steps_lim_ = steps_lim; };
    void setCostLim(TCost cost_lim)         { cost_lim_ = cost_lim; };
    void setTimeLim(time_t time_lim);

    // Parameters getters
    stop_t  getStopCond()    const { return stop_cond_; };
    count_t getStepsLim()    const { return steps_lim_; };
    TCost   getCostLim()     const { return cost_lim_; };
    time_t  getTimeLim()     const { return time_lim_; };

    // Other getters
    TSolution getBest()  const { return best_; };

    // Main method
    void run();

#ifdef SHC_DEBUG
	// Debugging methods
	void printParam() const;
    void printState() const;
    void printProgress() const;
#endif

protected:
	// Implement these pure virtual problem-dependent methods in derived class
	virtual TSolution generate() const = 0;
	virtual TCost     cost(TSolution const& solution) const = 0;
	virtual TSolution neighbour(TSolution const& solution) const = 0;
	virtual void      print(TSolution const& solution) const = 0;

	// SHC parameters
    stop_t  stop_cond_;
    count_t steps_lim_;
    TCost   cost_lim_;
    time_t  time_lim_;
 
    // SHC variables
    time_t    start_time_;
    count_t   step_;
    TSolution current_, new_, best_;
    TCost     current_cost_, new_cost_, best_cost_;

    // Stop condition checker
    boolean stopCondition();
};


template <class TCost, class TSolution>
void SHC<TCost, TSolution>::setTimeLim(time_t time_lim)
{
	if (time_lim >= 0)
	{
		time_lim_ = time_lim;
		return;
	};

#ifdef SHC_DEBUG
	Serial.println(F("Incorrect time limit!"));
#endif
};


template <class TCost, class TSolution>
void SHC<TCost, TSolution>::run()
{
    start_time_ = micros();

    // Calculate initial solution
    step_ = 0;

    current_      = generate();
    current_cost_ = cost(current_);

    best_         = current_;
    best_cost_    = current_cost_;

#ifdef SHC_DEBUG
        printProgress();
        //printState();
#endif

    // Main loop
    step_++;
    while (!stopCondition())
	{
		step_++;

		new_ = neighbour(current_);
		new_cost_ = cost(new_);

		if (new_cost_ <= current_cost_)
		{
			current_ = new_;
			current_cost_ = new_cost_;

			if (current_cost_ <= best_cost_)
			{
				best_ = current_;
				best_cost_ = current_cost_;
			}
		}

#ifdef SHC_DEBUG
		printProgress();
		printState();
#endif
	}
};


template <class TCost, class TSolution> 
boolean SHC<TCost, TSolution>::stopCondition()
{
    switch (stop_cond_) 
    {
        case STEPS_LIM: return (step_ > steps_lim_ ? true : false); 
        case COST_LIM:  return (best_cost_ <= cost_lim_ ? true : false);
        case TIME_LIM:  return ((micros() - start_time_) >= time_lim_ ? true : false);
    }
};


#ifdef SHC_DEBUG

template <class TCost, class TSolution> 
void SHC<TCost, TSolution>::printParam() const
{
    switch (stop_cond_)
    {
        case STEPS_LIM:     Serial.println(F("STEPS_LIM")); break;
        case COST_LIM:      Serial.println(F("COST_LIM")); break;
        case TIME_LIM:      Serial.println(F("TIME_LIM")); break;
    }

    Serial.print(F("steps_lim_ = "));
    Serial.println(steps_lim_);
    Serial.print(F("cost_lim_ = "));
    Serial.println(cost_lim_);
    Serial.print(F("time_lim_ = "));
    Serial.println(time_lim_);
}


template <class TCost, class TSolution> 
void SHC<TCost, TSolution>::printState() const
{
    Serial.println();
    Serial.print(F("Free RAM (bytes): "));
    Serial.println(freeRam());
    Serial.print(F("step_ = "));
    Serial.println(step_);
    Serial.println(F("current_:"));
    print(current_);
    Serial.println();
    Serial.print(F("current_cost_ = "));
    Serial.println(current_cost_);
    Serial.println(F("new_:"));
    print(new_);
    Serial.println();
    Serial.print(F("new_cost_ = "));
    Serial.println(new_cost_);
    Serial.println(F("best_:"));
    print(best_);
    Serial.println();
    Serial.print(F("best_cost_ = "));
    Serial.println(best_cost_);
}


template <class TCost, class TSolution> 
void SHC<TCost, TSolution>::printProgress() const
{
    Serial.print(F("step_ = "));
    Serial.println(step_);
    Serial.print(F("best_ = "));
    print(best_);
    Serial.println();
    Serial.print(F("cost_ = "));
    Serial.println(best_cost_);
}

#endif

#endif
